/**********************************************************************/
/*                                                                    */
/* Copyright (c) 2001 by Sun Microsystems, Inc.                       */
/* All rights reserved.                                               */
/*                                                                    */
/**********************************************************************/

/*
**
** This sample program demonstrates the use of the ECI to perform a Distributed
** Program Link (DPL) to a remote MTP system to determine the date and time.
**
** A single COBOL program is required on the MTP system concerned, called
** ECIEX1S, the source of which can be found in ../MTP/ECIEX1S.clt. This
** program should be compiled and installed on that MTP System.
**
** The MTP system must also be configured at accept TCP/IP connections. See
** the documentation for the MTP 'unikixtcp' process for details on how to
** do this.
**
** 
** NOTE: This sample is designed to illustrate various ECI programming
** techniques, and not as a mission-critical application. It therefore only
** performs minimal checking of return codes from system functions etc.     
*/                                                         

#ifdef _AIX
union wait;
#endif

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <sys/time.h>
#include <fcntl.h>
#ifdef LINUX
#include <ncurses.h>
#else
#include <curses.h>
#endif

#include <cics_eci.h>

/*
** The macros min and max are not always defined, so we define them here.
*/
#ifndef max
#define max(a,b) (a<b ? b : a )
#endif
#ifndef min
#define min(a,b) (a>b ? b : a )
#endif

/*
** The following defines are the return values from interuptable_getch()
** when either there is an error, or the FD supplied has data waiting on it.
*/
#define CHAR_INTERUPT	-2
#define CHAR_ERROR	-1

/*
** The following defines specify the positions on the screen that are used
** for the data.
*/
#define TOP_OF_LIST	6
#define STATUS_LINE	(LINES-5)
#define MAX_SYSTEMS	(STATUS_LINE-TOP_OF_LIST-1)

/*
** The following structure defines the data associated with a DPL in this
** environment.
*/
typedef enum _system_comand
{
    SYSTEM_COMMAND_NONE			=	0,
    SYSTEM_COMMAND_START,
    SYSTEM_COMMAND_STOP
} system_command_t;

typedef enum _system_state
{
    SYSTEM_STATE_IDLE			=	0,
    SYSTEM_STATE_DO_DPL,
    SYSTEM_STATE_ISSUED_DPL,
    SYSTEM_STATE_GETREPLY
} system_state_t;

typedef struct _dpl_info
{
    char		Alias[9];
    char		Date[9];
    char		Time[9];
    int			DPLCount;
    int			ReturnCode;
    char		AbendCode[4];
    system_command_t	Command;
    system_state_t	State;
    
    int			AskDateNext;
    cics_slong_t	LuwToken;
} dpl_info_t, *dpl_info_p;


static int		Client_FD=-1;
static cics_ushort_t	NumSystems=0;
static dpl_info_p	pSystems=NULL;
static int		TotalDPLCount=0;
static double		DPLsPerSecond=0.0;

/*
** This is a replacement to the curses getch() routine. The base curses
** version does not have the capability to wait for a char or some other
** event. This routine implements that via a select() call.
*/
static int interuptable_getch( int fd )
{
    fd_set              readfds;
    int                 ReturnChar;
    int                 rc;

    /*
    ** For this we use the select() call to wait for input on the
    ** curses stream, or the other stream. There is no 'legal' curses
    ** way to do this.
    */
    FD_ZERO(&readfds);
    FD_SET(0,&readfds);
    if ( fd >= 0 )
    {
	FD_SET(fd,&readfds);
    }
    rc = select( max(fd+1,1) , (fd_set *)&readfds,0,0,NULL);
    switch(rc)
    {
        case -1:
            ReturnChar = CHAR_ERROR;
            break;
        default:
	    if ( (fd >= 0) && FD_ISSET(fd,&readfds) )
	    {
		ReturnChar = CHAR_INTERUPT;
	    }
	    else
	    {
		ReturnChar=getch();
	    }
            break;
    }
    return(ReturnChar);
}


static int y,x;

static void SavePos( void )
{
    getyx(stdscr,y,x);
}

static void RestorePos( void )
{
    move(y,x);
}


static void DisplayBaseData( void )
{
    attrset(A_UNDERLINE);
    mvaddstr(0,24,"MTP Client ECI Sample - ECIEX2");
    attrset(A_NORMAL);

    mvaddstr(4,3,"System");
    mvaddstr(4,13,"Date");
    mvaddstr(4,23,"Time");
    mvaddstr(4,33,"DPL Count");
    mvaddstr(4,45,"RC");
    mvaddstr(4,55,"Abend");

    mvaddstr(STATUS_LINE+2,1,"Press Up/Down Arrow to select a system");
    mvaddstr(STATUS_LINE+3,1,"Press Space to start/stop getting the date/time");
    mvaddstr(STATUS_LINE+4,1,"Press 'q' to quit");
}


static void CallbackProc( cics_sshort_t MessageQualifier )
{
    pSystems[MessageQualifier].State = SYSTEM_STATE_GETREPLY;
}



static void DisplaySystemsData( int SystemsIndex )
{
    SavePos();
    move(TOP_OF_LIST+SystemsIndex,0);	clrtoeol();

    if ( pSystems[SystemsIndex].State != SYSTEM_STATE_IDLE )
    {
	mvaddstr(TOP_OF_LIST+SystemsIndex,1,"*");
    }
    mvaddstr(TOP_OF_LIST+SystemsIndex,3,pSystems[SystemsIndex].Alias);
    mvprintw(TOP_OF_LIST+SystemsIndex,13,"%-8.8s",pSystems[SystemsIndex].Date);
    mvprintw(TOP_OF_LIST+SystemsIndex,23,"%-8.8s",pSystems[SystemsIndex].Time);
    mvprintw(TOP_OF_LIST+SystemsIndex,33,"%d",pSystems[SystemsIndex].DPLCount);
    mvprintw(TOP_OF_LIST+SystemsIndex,45,"%d",pSystems[SystemsIndex].ReturnCode);
    mvprintw(TOP_OF_LIST+SystemsIndex,55,"%-4.4s",pSystems[SystemsIndex].AbendCode);
    RestorePos();
}



static void PerformDPL( int SystemsIndex )
{
    /*
    ** This function is responsible for performing the actual distributed
    ** program link (DPL) calls to the remote MTP system.
    ** Depending on the state of the conversation, it will either ask MTP
    ** for the Date, or the Time.
    **/
    ECI_PARMS       EciParms;
    unsigned char   *AskTime	      = (unsigned char *)"ASKTIME:";
    unsigned char   *AskDate	      = (unsigned char *)"ASKDATE/";
    unsigned char   EciEx1Prog[]      = "ECIEX1S "; 
    unsigned char   EciUserID[]       = "        "; 
    unsigned char   EciPassword[]     = "        "; 
    cics_sshort_t   RetCode;


    /*
    ** Clean out the ECI parameter block.
    */
    memset(&EciParms,0,sizeof(ECI_PARMS));

    /*
    ** Set the ECI parameter block to our required values
    */

    EciParms.eci_call_type		= ECI_ASYNC;
    memcpy(EciParms.eci_program_name,  	EciEx1Prog,8);
    memcpy(EciParms.eci_userid,        	EciUserID,8);
    memcpy(EciParms.eci_password,      	EciPassword,8);
    
    if ( pSystems[SystemsIndex].AskDateNext )
    {
	EciParms.eci_commarea         	= AskDate;
	EciParms.eci_extend_mode	= ECI_NO_EXTEND;
    }
    else
    {
	if (pSystems[SystemsIndex].Command == SYSTEM_COMMAND_STOP)
	{
	    /*
	    ** The user has pressed 'stop' so do not start a new UOW
	    */
	    pSystems[SystemsIndex].Command = SYSTEM_COMMAND_NONE;
	    pSystems[SystemsIndex].State   = SYSTEM_STATE_IDLE;
	    return;
	}
	EciParms.eci_commarea         	= AskTime;
	EciParms.eci_extend_mode	= ECI_EXTENDED;
    }
    
    EciParms.eci_commarea_length	= 8;
    
    
    EciParms.eci_luw_token		= pSystems[SystemsIndex].LuwToken;
    
    EciParms.eci_version                = ECI_VERSION_1;
    EciParms.eci_message_qualifier      = SystemsIndex;
    EciParms.eci_callback	        = CallbackProc;
    memcpy(EciParms.eci_system_name,    pSystems[SystemsIndex].Alias,8);
	    
    pSystems[SystemsIndex].ReturnCode = CICS_ExternalCall(&EciParms);
    pSystems[SystemsIndex].LuwToken	= EciParms.eci_luw_token;
    
    if (pSystems[SystemsIndex].ReturnCode == ECI_NO_ERROR)
    {
	pSystems[SystemsIndex].State = SYSTEM_STATE_ISSUED_DPL;
	return;
    }
    else
    {
	pSystems[SystemsIndex].LuwToken= ECI_LUW_NEW;
	pSystems[SystemsIndex].State = SYSTEM_STATE_IDLE;
    }
    pSystems[SystemsIndex].ReturnCode++;
	    
    return;
}




static void PerformGetReply( int Index )
{
    ECI_PARMS       EciParms;
    char	    CommArea[8];
    unsigned char   EciEx1Prog[]      = "ECIEX1S "; 
    unsigned char   EciUserID[]       = "        "; 
    unsigned char   EciPassword[]     = "        "; 


    /*
    ** Clean out the ECI parameter block.
    */
    memset(&EciParms,0,sizeof(ECI_PARMS));

    /*
    ** Set the ECI parameter block to our required values
    */

    EciParms.eci_call_type		= ECI_GET_SPECIFIC_REPLY;
    memcpy(EciParms.eci_program_name,  	EciEx1Prog,8);
    memcpy(EciParms.eci_userid,        	EciUserID,8);
    memcpy(EciParms.eci_password,      	EciPassword,8);
    
    EciParms.eci_commarea         	= CommArea;
    
    EciParms.eci_commarea_length	= 8;
    
    
    EciParms.eci_luw_token		= ECI_LUW_NEW;
    
    EciParms.eci_version                = ECI_VERSION_1;

    EciParms.eci_message_qualifier	= Index;
	    
    pSystems[Index].ReturnCode = CICS_ExternalCall(&EciParms);
    /*pSystems[Index].RequestInProgress = 0;
    pSystems[Index].ReplyReady=0;*/
    memcpy(pSystems[Index].AbendCode,EciParms.eci_abend_code,4);
    
    if (pSystems[Index].ReturnCode == ECI_NO_ERROR)
    {
	if ( pSystems[Index].AskDateNext )
	{
	    memcpy( pSystems[Index].Date,
		    CommArea,
		    8);
	}
	else
	{
	    memcpy( pSystems[Index].Time,
		    CommArea,
		    8);
	}
	pSystems[Index].DPLCount++;
	TotalDPLCount++;

	pSystems[Index].State = SYSTEM_STATE_DO_DPL;

	/*
	** Toggle the AskDateNext flag
	*/
	pSystems[Index].AskDateNext = 1 - pSystems[Index].AskDateNext;
    }
    else
    {
	/*
	** Something went wrong, so stop doing work
	** to this end-system. We assume that the Unit-Of-Work has
	** been terminated
	*/
	pSystems[Index].AskDateNext = 0;
	pSystems[Index].Command = SYSTEM_COMMAND_STOP;
	pSystems[Index].State = SYSTEM_STATE_IDLE;
    }


    DisplaySystemsData( Index );
	    
    return;
}

static cics_sshort_t PopulateSystemsList( void )
{
    cics_sshort_t	clirc;
    CICS_EciSystem_t	*pTmp;
    int			i;

    /*
    ** Initially call CICS_EciListSystems() with 0 as the number of systems.
    ** This causes the ECI to return the number of systems that are defined.
    ** We then use the number of systems returned to identify how many systems
    ** to ask for and how many to display.
    */
    clirc = CICS_EciListSystems( NULL, &NumSystems , NULL );
    if ( clirc == ECI_ERR_MORE_SYSTEMS )
    {
	NumSystems = min((int)NumSystems,(int)MAX_SYSTEMS);
	pSystems = (dpl_info_p)malloc(sizeof(dpl_info_t)*NumSystems);
	memset(pSystems,'\0',sizeof(dpl_info_t)*NumSystems);
	pTmp = (CICS_EciSystem_t *)
			malloc(sizeof(CICS_EciSystem_t)*NumSystems);

	clirc = CICS_EciListSystems( NULL , &NumSystems , pTmp );
	if ( ( clirc == ECI_NO_ERROR ) ||
	     ( clirc == ECI_ERR_MORE_SYSTEMS ) )
	{
	    clirc = ECI_NO_ERROR;
	    NumSystems = min((int)NumSystems,(int)MAX_SYSTEMS);
	    for ( i=0;i<(int)NumSystems;i++ )
	    {
		memcpy(pSystems[i].Alias,pTmp[i].SystemName,8);
	    }
	}
    }
    return clirc;
}


static void PerformChangeRunningState( void )
{
    int x,y;

    getyx(stdscr,y,x);

    if ( pSystems[y-TOP_OF_LIST].State != SYSTEM_STATE_IDLE )
    {
	pSystems[y-TOP_OF_LIST].Command = SYSTEM_COMMAND_STOP;
    }
    else
    {
	pSystems[y-TOP_OF_LIST].Command = SYSTEM_COMMAND_START;
    }

    DisplaySystemsData(y-TOP_OF_LIST);

    move(y,x);
}


/*
** This function is called to potentially update the screen with the DPLS per
** second. It looks to see if 20 seconds have gone by, and uif it has, the
** screen gets updated.
*/
static void UpdateDPLsPerSecond( void )
{
    static time_t	SavedTime=0;
    time_t		NewTime;
    static int		SavedDPLCount;

    if ( SavedTime == 0 )
    {
	SavedTime=time(NULL);
	SavedDPLCount = TotalDPLCount;
    }
    else
    {
	NewTime = time(NULL);
	if ( NewTime - SavedTime > 20 )
	{
	    DPLsPerSecond = ((double)(TotalDPLCount-SavedDPLCount))/
				((double)(NewTime-SavedTime));
	    mvprintw(STATUS_LINE,3,"Total DPLs per second = %f",DPLsPerSecond);
	    SavedTime = NewTime;
	    SavedDPLCount = TotalDPLCount;
	}
    }
}

static void ProcessUpKey( void )
{
    int y,x;
    getyx(stdscr,y,x);
    if ( y>TOP_OF_LIST  )
    {
	move(y-1,x);
    }
    else
    {
	beep();
    }
    refresh();
}

static void ProcessDownKey( void )
{
    int y,x;
    getyx(stdscr,y,x);
    if ( y<(int)(NumSystems+TOP_OF_LIST-1)  )
    {
	move(y+1,x);
    }
    else
    {
	beep();
    }
    refresh();
}

int main( int ArgC , char **ArgV )
{
    int			Done=0;
    int			c;	/* Character obtained by getch()	*/
    int			i;
    cics_sshort_t	clirc;
    int			Restart = 1;
    int			Dummy;

    /*
    ** Initialize curses.
    */
    initscr();
    crmode();
    noecho();
    keypad(stdscr,TRUE);
    nodelay(stdscr,TRUE);

    getch();	/* Dummy getch() to make interuptable_getch() work! */

    /*
    ** Get a list of the systems available, and populate our structures.
    */
    if ( (clirc = PopulateSystemsList()) != ECI_NO_ERROR )
    {
	clear();
	refresh();
	endwin();
	fprintf(stderr,"Attach to MTP Client Failed %d\n",clirc);
	exit(1);
    }

    /*
    ** Display the framework of our screen.
    */
    DisplayBaseData();

    /*
    ** Display the data associated with each of the systems.
    */
    for ( i=0;i<(int)NumSystems;i++ )
    {
	DisplaySystemsData(i);
    }

    move(TOP_OF_LIST,0);
    refresh();

    /*
    ** Find the Sun MTP Client ECI FD to select on.
    */
    KixCli_QueryFD(&Client_FD);

    /*
    ** Now the main loop. Keep on going until the user hits a QUIT key.
    */
    while( !Done )
    {

	cics_sshort_t	ecirc;

	Restart = 0;

	SavePos();

	/*
	** First call into the ECI to enable the callbacks to be performed.
	*/
	ecirc = KixCli_QueryFD(&Dummy);

	/*
	** First test to see if there are any outstanding replies
	*/
	for ( i=0;i<(int)NumSystems;i++ )
	{
	    if ( pSystems[i].State == SYSTEM_STATE_GETREPLY )
	    {
		PerformGetReply(i);
		Restart = 1;
	    }
	}
	
	/*
	** A GetReply call can itself  lead to new callbacks, so go
	** round the loop again to check for this
	*/
	if (Restart)
	{
	    continue;
	}

	/*
	** Once we get here, there are currently NO replies outstanding,
	** so dispatch any outstanding DPLs
	*/
	for ( i=0;i<(int)NumSystems;i++ )
	{
	    if ( ( pSystems[i].State == SYSTEM_STATE_DO_DPL ) ||
                 ( ( pSystems[i].State == SYSTEM_STATE_IDLE ) &&
		   ( pSystems[i].Command == SYSTEM_COMMAND_START )))
	    {
		PerformDPL(i);
		DisplaySystemsData(i);
		Restart = 1;
	    }
	}

	/*
	** A DPL call can lead to new callbacks, so go round the loop
	** again to check for this
	*/
	if (Restart)
	{
	    continue;
	}

	/*
	** Update the DPLs per second value if we have reached a decent time
	** interval.
	*/
	UpdateDPLsPerSecond();
	RestorePos();
	refresh();

	/*
	** If we get here, there is nothng for us to do but go on the select
	*/
	c = interuptable_getch(Client_FD);

	switch( c )
	{
	    case CHAR_ERROR:
		mvprintw(STATUS_LINE,2,"The MTP Client has been terminated");
		refresh();
		break;

	    case CHAR_INTERUPT:
		/*
		** Just fall through, and thus round to the GetReply code
		*/
		break;

	    case ' ':
		if ( Client_FD >= 0 )
		{
		    SavePos();
		    PerformChangeRunningState();
		    RestorePos();
		    refresh();
		}
		else
		{
		    beep();
		}
		break;
	    case 'Q':
	    case 'q':
		Done = 1;
		break;
	    case KEY_DOWN:
		if ( Client_FD >= 0 )	ProcessDownKey();
		else				beep();
		break;
	    case KEY_UP:
		if ( Client_FD >= 0 )	ProcessUpKey();
		else				beep();
		break;
	    default:
		beep();
		break;
	}


    }
    clear();
    refresh();
    endwin();

    /*
    ** If we got storage for the systems structures, then free it.
    */
    if ( pSystems )
    {
	free(pSystems);
    }
    exit(0);
}
