/**********************************************************************/
/*                                                                    */
/* Copyright (c) 2001 by Sun Microsystems, Inc.                       */
/* All rights reserved.                                               */
/*                                                                    */
/**********************************************************************/

/*
**
** This sample program demonstrates the use of the EPI to execute transactions
** on a remote MTP system to determine the date and time.
**
** A single COBOL program is required on the MTP system concerned, called
** ECIEX1S, the source of which can be found in ..\MTP\ECIEX1S.clt. This
** program should be compiled and installed on that MTP System.
**
** The MTP system must also be configured at accept TCP/IP connections. See
** the documentation for the MTP 'unikixtcp' process for details on how to
** do this.
**
** 
** NOTE: This sample is designed to illustrate various EPI programming
** techniques, and not as a mission-critical application. It therefore only
** performs minimal checking of return codes from system functions etc.     
*/                                                         

#ifdef _AIX
union wait;
#endif

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <stdarg.h>
#include <sys/time.h>
#include <fcntl.h>
#ifdef LINUX
#include <ncurses.h>
#else
#include <curses.h>
#endif

#include <cics_epi.h>

/*
** The macros min and max are not always defined, so we define them here.
*/
#ifndef max
#define max(a,b) (a<b ? b : a )
#endif
#ifndef min
#define min(a,b) (a>b ? b : a )
#endif

/*
** The following defines are the return values from interuptable_getch()
** when either there is an error, or the FD supplied has data waiting on it.
*/
#define CHAR_INTERUPT	-2
#define CHAR_ERROR	-1

/*
** The following defines specify the positions on the screen that are used
** for the data.
*/
#define TOP_OF_LIST	6
#define STATUS_LINE	(LINES-5)
#define MAX_SYSTEMS	(STATUS_LINE-TOP_OF_LIST-1)

/*
** The following defines define the datastreams to be sent when a specific
** key is required.
*/
#define DATASTREAM_ENTER	(cics_ubyte_t *)"'  "
#define DATASTREAM_PF3		(cics_ubyte_t *)"3  "

/*
** The following structure defines the data associated with a request in this
** environment.
*/
typedef struct _dpl_info
{
    char		Alias[9];
    char		NetName[9];
    char		Date[9];
    char		Time[9];
    int			DPLCount;
    char		AbendCode[4];
    int			Running;
    int			RequestInProgress;
    int			EventReady;
    int			AskDateNext;
    cics_ushort_t	TermIndex;
    cics_ushort_t	ReturnCode;
} dpl_info_t, *dpl_info_p;

static int		Client_FD=-1;
static cics_ushort_t	NumSystems=0;
static dpl_info_p	pSystems=NULL;
static int		TotalDPLCount=0;
static double		DPLsPerSecond=0.0;

/*
** This is a replacement to the curses getch() routine. The base curses
** version does not have the capability to wait for a char or some other
** event. This routine impliments that via a select() call.
*/
static int interuptable_getch( int fd )
{
    fd_set              readfds;
    int                 ReturnChar;
    int                 rc;

    /*
    ** For this we use the select() call to wait for input on the
    ** curses stream, or the other stream. There is no 'legal' curses
    ** way to do this.
    */
    FD_ZERO(&readfds);
    FD_SET(0,&readfds);
    if ( fd >= 0 )
    {
	FD_SET(fd,&readfds);
    }
    rc = select( max(fd+1,1) , (fd_set *)&readfds,0,0,NULL);
    switch(rc)
    {
        case -1:
            ReturnChar = CHAR_ERROR;
            break;
        default:
	    if ( (fd >= 0) && FD_ISSET(fd,&readfds) )
	    {
		ReturnChar = CHAR_INTERUPT;
	    }
	    else
	    {
		ReturnChar=getch();
	    }
            break;
    }
    return(ReturnChar);
}


static int y,x;

static void SavePos( void )
{
    getyx(stdscr,y,x);
}

static void RestorePos( void )
{
    move(y,x);
}


static void DisplayBaseData( void )
{
    attrset(A_UNDERLINE);
    mvaddstr(0,24,"MTP Client EPI Sample - EPIEX2");
    attrset(A_NORMAL);

    mvaddstr(4,3,"System");
    mvaddstr(4,13,"Date");
    mvaddstr(4,23,"Time");
    mvaddstr(4,33,"Tran Count");
    mvaddstr(4,45,"RC");
    mvaddstr(4,55,"Abend");
    mvaddstr(4,65,"NetName");
    mvaddstr(4,75,"Index");

    mvaddstr(STATUS_LINE+2,1,"Press Up/Down Arrow to select a system");
    mvaddstr(STATUS_LINE+3,1,"Press Space to start/stop getting the date/time");
    mvaddstr(STATUS_LINE+4,1,"Press 'q' to quit");
}


static void CallbackProc( cics_ushort_t TermIndex )
{
    int		i;

    for ( i=0;i<NumSystems;i++ )
    {
	if ( pSystems[i].TermIndex == TermIndex )
	{
	    pSystems[i].EventReady=1;
	}
    }
}



static void DisplaySystemsData( int SystemsIndex )
{
    move(TOP_OF_LIST+SystemsIndex,0);	clrtoeol();

    if ( pSystems[SystemsIndex].Running )
    {
	mvaddstr(TOP_OF_LIST+SystemsIndex,1,"*");
    }
    mvaddstr(TOP_OF_LIST+SystemsIndex,3,pSystems[SystemsIndex].Alias);
    mvprintw(TOP_OF_LIST+SystemsIndex,13,"%-8.8s",pSystems[SystemsIndex].Date);
    mvprintw(TOP_OF_LIST+SystemsIndex,23,"%-8.8s",pSystems[SystemsIndex].Time);
    mvprintw(TOP_OF_LIST+SystemsIndex,33,"%d",pSystems[SystemsIndex].DPLCount);
    mvprintw(TOP_OF_LIST+SystemsIndex,45,"%d",pSystems[SystemsIndex].ReturnCode);
    mvprintw(TOP_OF_LIST+SystemsIndex,55,"%-4.4s",pSystems[SystemsIndex].AbendCode);
    mvprintw(TOP_OF_LIST+SystemsIndex,65,"%-8.8s",pSystems[SystemsIndex].NetName);
    if ( pSystems[SystemsIndex].TermIndex == CICS_EPI_TERM_INDEX_NONE )
    {
    mvprintw(TOP_OF_LIST+SystemsIndex,75,"     ");
    }
    else
    {
    mvprintw(TOP_OF_LIST+SystemsIndex,75,"%d",pSystems[SystemsIndex].TermIndex);
    }
}


static void StartTransaction( int SystemsIndex )
{
    cics_ushort_t	epirc;

    epirc = CICS_EpiStartTran( pSystems[SystemsIndex].TermIndex,
			       "EPX1",
			       DATASTREAM_ENTER,
			       3 );
    pSystems[SystemsIndex].ReturnCode = epirc;

    if ( epirc != CICS_EPI_NORMAL )
    {
	pSystems[SystemsIndex].Running = 0;
    }

}


static void EndTranPF3( int SystemsIndex )
{
    cics_ushort_t	epirc;

    pSystems[SystemsIndex].DPLCount++;
    TotalDPLCount++;

    epirc = CICS_EpiStartTran( pSystems[SystemsIndex].TermIndex,
			       "EPX1",
			       DATASTREAM_PF3,
			       3 );
    pSystems[SystemsIndex].ReturnCode = epirc;

    if ( epirc != CICS_EPI_NORMAL )
    {
	pSystems[SystemsIndex].Running = 0;
    }
}

static void EndTranEnter( int SystemsIndex )
{
    cics_ushort_t	epirc;

    pSystems[SystemsIndex].DPLCount++;
    TotalDPLCount++;

    epirc = CICS_EpiStartTran( pSystems[SystemsIndex].TermIndex,
			       "EPX1",
			       DATASTREAM_ENTER,
			       3 );
    pSystems[SystemsIndex].ReturnCode = epirc;

    if ( epirc != CICS_EPI_NORMAL )
    {
	pSystems[SystemsIndex].Running = 0;
    }
}

static void ConversePF3( int SystemsIndex )
{
    cics_ushort_t	epirc;

    pSystems[SystemsIndex].DPLCount++;
    TotalDPLCount++;

    epirc = CICS_EpiReply( pSystems[SystemsIndex].TermIndex,
			       DATASTREAM_PF3,
			       3 );
    pSystems[SystemsIndex].ReturnCode = epirc;

    if ( epirc != CICS_EPI_NORMAL )
    {
	pSystems[SystemsIndex].Running = 0;
    }
}

static void ConverseEnter( int SystemsIndex )
{
    cics_ushort_t	epirc;

    pSystems[SystemsIndex].DPLCount++;
    TotalDPLCount++;

    epirc = CICS_EpiReply( pSystems[SystemsIndex].TermIndex,
			       DATASTREAM_ENTER,
			       3 );
    pSystems[SystemsIndex].ReturnCode = epirc;

    if ( epirc != CICS_EPI_NORMAL )
    {
	pSystems[SystemsIndex].Running = 0;
    }
}

static void PerformDPL( int SystemsIndex )
{
    cics_ushort_t	epirc;

}




static void PerformGetEvent( int SystemIndex )
{
    cics_ushort_t	clirc=CICS_EPI_ERR_MORE_EVENTS;
    CICS_EpiEventData_t	Event;
    cics_ubyte_t	DataArea[7000];

    pSystems[SystemIndex].ReturnCode = CICS_EPI_ERR_MORE_EVENTS;
    while ( pSystems[SystemIndex].ReturnCode == CICS_EPI_ERR_MORE_EVENTS )
    {
	memset(DataArea,'\0',7000);
	Event.Data = DataArea;
	Event.Size = 7000;
	pSystems[SystemIndex].ReturnCode = CICS_EpiGetEvent(
				  pSystems[SystemIndex].TermIndex,
				  CICS_EPI_NOWAIT,
				  &Event );
	if ( ( pSystems[SystemIndex].ReturnCode == CICS_EPI_ERR_MORE_EVENTS ) ||
	     ( pSystems[SystemIndex].ReturnCode == CICS_EPI_NORMAL ) )

	{
	    switch ( Event.Event )
	    {
		case CICS_EPI_EVENT_SEND:
		    if ( Event.Size > 0x8a )
		    {
			if (memcmp(&(DataArea[0x13]),"MAP1",4) == 0)
			{
			    memcpy(pSystems[SystemIndex].Date,&(DataArea[0x82]),8);
			}
			else if (memcmp(&(DataArea[0x0F]),"MAP1",4) == 0)
			{
			    memcpy(pSystems[SystemIndex].Date,&(DataArea[0x7E]),8);
			}
			else if (memcmp(&(DataArea[0x0F]),"MAP2",4) == 0)
			{
			    memcpy(pSystems[SystemIndex].Time,&(DataArea[0x7E]),8);
			}
			else if (memcmp(&(DataArea[0x13]),"MAP2",4) == 0)
			{
			    memcpy(pSystems[SystemIndex].Time,&(DataArea[0x82]),8);
			}
			else
			{
			}
		    }
		    break;
		case CICS_EPI_EVENT_CONVERSE:
		    if ( pSystems[SystemIndex].Running )
		    {
			ConverseEnter(SystemIndex);
		    }
		    else
		    {
			ConversePF3(SystemIndex);
		    }
		    break;
		case CICS_EPI_EVENT_END_TRAN:
		    if ( Event.TransId[0] != '\0' )
		    {
			if ( pSystems[SystemIndex].Running )
			{
			    EndTranEnter(SystemIndex);
			}
			else
			{
			    EndTranPF3(SystemIndex);
			}
		    }
		    else
		    {
			pSystems[SystemIndex].ReturnCode = CICS_EpiDelTerminal(
				pSystems[SystemIndex].TermIndex );
		    }
		    break;
		case CICS_EPI_EVENT_END_TERM:
		    pSystems[SystemIndex].TermIndex = CICS_EPI_TERM_INDEX_NONE;
		    memset(pSystems[SystemIndex].NetName,'\0',9);
		    break;
		default:
		    break;
	    }
	}
    }
    pSystems[SystemIndex].EventReady = 0;

    DisplaySystemsData( SystemIndex );
	    
    return;
}






static cics_ushort_t PopulateSystemsList( void )
{
    cics_ushort_t	clirc;
    CICS_EpiSystem_t	*pTmp;
    int			i;

    clirc = CICS_EpiListSystems( NULL, &NumSystems , NULL );
    if ( clirc == CICS_EPI_ERR_MORE_SYSTEMS )
    {
	NumSystems = min((int)NumSystems,(int)MAX_SYSTEMS);
	pSystems = (dpl_info_p)malloc(sizeof(dpl_info_t)*NumSystems);
	memset(pSystems,'\0',sizeof(dpl_info_t)*NumSystems);
	pTmp = (CICS_EpiSystem_t *)
			malloc(sizeof(CICS_EpiSystem_t)*NumSystems);
	clirc = CICS_EpiListSystems( NULL , &NumSystems , pTmp );
	if ( ( clirc == CICS_EPI_NORMAL ) ||
	     ( clirc == CICS_EPI_ERR_MORE_SYSTEMS ) )
	{
	    clirc = CICS_EPI_NORMAL;
	    NumSystems = min((int)NumSystems,(int)MAX_SYSTEMS);
	    for ( i=0;i<(int)NumSystems;i++ )
	    {
		memcpy(pSystems[i].Alias,pTmp[i].SystemName,8);
		pSystems[i].TermIndex = CICS_EPI_TERM_INDEX_NONE;
	    }
	}
    }
    return clirc;
}

static void UpdateDPLsPerSecond( void )
{
    static time_t	SavedTime=0;
    time_t		NewTime;
    static int		SavedDPLCount;

    if ( SavedTime == 0 )
    {
	SavedTime=time(NULL);
	SavedDPLCount = TotalDPLCount;
    }
    else
    {
	NewTime = time(NULL);
	if ( NewTime - SavedTime > 20 )
	{
	    DPLsPerSecond = ((double)(TotalDPLCount-SavedDPLCount))/
				((double)(NewTime-SavedTime));
	    mvprintw(STATUS_LINE,2,"Total Trans per second = %f",DPLsPerSecond);
	    SavedTime = NewTime;
	    SavedDPLCount = TotalDPLCount;
	}
    }
}

static void ProcessInterupt( void )
{
    int i;
    cics_sshort_t	ecirc;
    int			Dummy;
    int			ThereWereSome = 1;

    SavePos();

    /*
    ** First call into the ECI to enable the callbacks to be performed.
    */
    ecirc = KixCli_QueryFD(&Dummy);

    /*
    ** While any of the DPLs has a reply ready, perform the GetReply()
    */
    while ( ThereWereSome )
    {
	ThereWereSome=0;
	for ( i=0;i<(int)NumSystems;i++ )
	{
	    if ( pSystems[i].EventReady )
	    {
		PerformGetEvent(i);
		ThereWereSome=1;
	    }
	}
    }

    /*
    ** Update the DPLs per second value if we have reached a decent time
    ** interval.
    */
    UpdateDPLsPerSecond();

    /*
    ** Now set off the new DPLs
    */
    for ( i=0;i<(int)NumSystems;i++ )
    {
	if ( pSystems[i].Running && !pSystems[i].RequestInProgress )
	{
	    PerformDPL(i);
	    DisplaySystemsData(i);
	}
    }
    RestorePos();
    refresh();
}

static void PerformChangeRunningState( void )
{
    int x,y;
    cics_ushort_t	epirc;
    CICS_EpiDetails_t	Details;

    getyx(stdscr,y,x);
    if ( pSystems[y-TOP_OF_LIST].Running )
    {
	pSystems[y-TOP_OF_LIST].Running=0;
    }
    else
    {
	pSystems[y-TOP_OF_LIST].Running=1;
	if ( pSystems[y-TOP_OF_LIST].TermIndex == CICS_EPI_TERM_INDEX_NONE )
	{
	    epirc = CICS_EpiAddTerminal( NULL,
					 pSystems[y-TOP_OF_LIST].Alias,
					 NULL,
					 NULL,
					 CallbackProc,
					 &Details,
					 &pSystems[y-TOP_OF_LIST].TermIndex );
	    pSystems[y-TOP_OF_LIST].ReturnCode = epirc;
	    if ( epirc != CICS_EPI_NORMAL )
	    {
		pSystems[y-TOP_OF_LIST].Running=0;
	    }
	    else
	    {
		memcpy(pSystems[y-TOP_OF_LIST].NetName,Details.NetName,9);
		StartTransaction(y-TOP_OF_LIST);
	    }
	    ProcessInterupt();
	}
    }
    DisplaySystemsData(y-TOP_OF_LIST);

    move(y,x);
}

static void ProcessUpKey( void )
{
    int y,x;
    getyx(stdscr,y,x);
    if ( y>TOP_OF_LIST  )
    {
	move(y-1,x);
    }
    else
    {
	beep();
    }
    refresh();
}

static void ProcessDownKey( void )
{
    int y,x;
    getyx(stdscr,y,x);
    if ( y<(int)(NumSystems+TOP_OF_LIST-1)  )
    {
	move(y+1,x);
    }
    else
    {
	beep();
    }
    refresh();
}

int main( int ArgC , char **ArgV )
{
    int			Done=0;
    int			c;	/* Character obtained by getch()	*/
    int			i;
    cics_ushort_t	clirc;

    /*
    ** Initialize curses.
    */
    initscr();
    crmode();
    noecho();
    keypad(stdscr,TRUE);
    nodelay(stdscr,TRUE);

    getch();	/* Dummy getch() to make interuptable_getch() work! */

    /*
    ** Initialize the EPI
    */
    clirc = CICS_EpiInitialize( CICS_EPI_VERSION_101 );
    if ( clirc != CICS_EPI_NORMAL )
    {
	clear();
	refresh();
	endwin();
	fprintf(stderr,"Attach to MTP Client Failed %d\n",clirc);
	exit(1);
    }

    /*
    ** Get a list of the systems available, and populate our structures.
    */
    if ( (clirc = PopulateSystemsList()) != CICS_EPI_NORMAL )
    {
	clear();
	refresh();
	endwin();
	fprintf(stderr,"Attach to MTP Client Failed %d\n",clirc);
	exit(1);
    }

    /*
    ** Display the framework of our screen.
    */
    DisplayBaseData();

    /*
    ** Display the data associated with each of the systems.
    */
    for ( i=0;i<(int)NumSystems;i++ )
    {
	DisplaySystemsData(i);
    }

    move(TOP_OF_LIST,0);
    refresh();

    /*
    ** Find the Sun MTP Client ECI FD to select on.
    */
    KixCli_QueryFD(&Client_FD);

    /*
    ** Now the main loop. Keep on going until the user hits a QUIT key.
    */
    while( !Done )
    {
	c = interuptable_getch(Client_FD);
	switch( c )
	{
	    case CHAR_ERROR:
		mvprintw(STATUS_LINE,2,"The MTP Client has been terminated");
		KixCli_QueryFD(&Client_FD);
		refresh();
		break;
	    case CHAR_INTERUPT:
		ProcessInterupt();
		break;
	    case ' ':
		if ( Client_FD >= 0 )
		{
		    SavePos();
		    PerformChangeRunningState();
		    RestorePos();
		    refresh();
		}
		else
		{
		    beep();
		}
		break;
	    case 'Q':
	    case 'q':
		Done = 1;
		break;
	    case KEY_DOWN:
		if ( Client_FD >= 0 )	ProcessDownKey();
		else				beep();
		break;
	    case KEY_UP:
		if ( Client_FD >= 0 )	ProcessUpKey();
		else				beep();
		break;
	    default:
		beep();
		break;
	}
    }
    clear();
    refresh();
    endwin();

    /*
    ** If we got storage for the systems structures, then free it.
    */
    if ( pSystems )
    {
	free(pSystems);
    }

    clirc = CICS_EpiTerminate();

    exit(0);
}
