/**********************************************************************/
/*                                                                    */
/* Copyright (c) 2002 by Sun Microsystems, Inc.                       */
/* All rights reserved.                                               */
/*                                                                    */
/**********************************************************************/

import com.sun.emp.cci.eci.ECIInteractionSpec;
import com.sun.emp.cci.eci.ECIManagedConnectionFactory;
import com.sun.emp.cci.eci.ECITransactionAbendException;
import com.sun.emp.cci.eci.GenericRecord;

import java.io.BufferedReader;
import java.io.InputStreamReader;
import java.io.IOException;

import javax.resource.cci.Connection;
import javax.resource.cci.ConnectionFactory;
import javax.resource.cci.Interaction;
import javax.resource.cci.InteractionSpec;

import javax.resource.ResourceException;

/**
 * an example unmanaged program using the MTP ECI Resource Adapter.
 * <p>This example obtains the data and time for an MTP system using
 * the MTP ECI Resource Adapter.
 */
public class Sample1 {

    private ConnectionFactory	cf;

    public Sample1(String host, int port, String regionName )
    						throws ResourceException {

	//
	// Create the managed connection factory and configure its
	// properties (as supplied as arguiments to this method)
	//
	ECIManagedConnectionFactory emcf = new ECIManagedConnectionFactory();
	emcf.setNameServiceHost(host);
	emcf.setNameServicePort(new Integer(port));
	emcf.setRegionName(regionName);

	//
	// Use the managed connection factory to create a connection factory
	//
	cf  = (ConnectionFactory)emcf.createConnectionFactory();
    }

    /**
     * Actually perform a query agains the MTP system.
     * <p>This method uses the <code>ECIEX1S</code> program on the server
     * to obtain information. There are 2 possible queries, one for the
     * date and one for the time. The methodology used for these is identical,
     * so this single method can be used for both.
     * @param query The query string
     * @throws javax.resource.ResourceException if anything goes wrong.
     */
    private String performQuery(String query) throws ResourceException {

	//
	// Create and configure the interaction spec. This tells the
	// server what MTP program to run.
	//
	ECIInteractionSpec eis = new ECIInteractionSpec();

	eis.setInteractionVerb(InteractionSpec.SYNC_SEND_RECEIVE);
	eis.setFunctionName("ECIEX1S");

	//
	// Create a record to holding the input commarea. Note that this
	// will also hold the resultant commarea after the program is
	// actually executed.
	//
	GenericRecord gr = new GenericRecord(query.getBytes());

	Connection c   = null;

	try {
	    //
	    // Get a connection from the connection factory.
	    // This will be 'pooled' by the default connection manager
	    //
	    c  = cf.getConnection();

	    //
	    // Finally execute the program by creating an interaction and
	    // executing the program within it.
	    //
	    Interaction ia = c.createInteraction();
	    ia.execute(eis,gr,gr);

	    //
	    // Prepare the data to return
	    //
	    String date =  new String(gr.getBytes());

	    return date;
        } finally {
	    //
	    // Whether we were successful or not, we must close the connection.
	    // This allows connection pooling to occur correctly and it allows
	    // resources to be released.
	    //
            if (c != null) {
                try {
                    c.close();
                } catch (ResourceException re) {
		    //
		    // Dont worry if this fails
		    //
                }
            }
        }
    }

    /**
     * Query the date from the MTP System
     * @return The date in the form YY/MM/DD
     * @throws javax.resource.ResourceException if something fails
     */
    public String getDate() throws ResourceException {
	return performQuery("ASKDATE/");
    }


    /**
     * Query the time from the MTP System
     * @return The time in the form HH:MM:SS
     * @throws javax.resource.ResourceException if something fails
     */
    public String getTime() throws ResourceException {
	return performQuery("ASKTIME:");
    }

    public static void main(String[] args) {
	//
	// Get hold of the required arguments.
	//
	if ( args.length == 0 ) {
	    try {
		args = getArgsFromStdIn();
	    } catch ( IOException e ) {
		usage();
		System.exit(1);
	    }
	}
	if ( args.length<3 ) {
	    usage();
	    System.exit(1);
	}
	String host = args[0];
	int port=0;
	try {
	    port = Integer.parseInt(args[1]);
	} catch ( NumberFormatException e ) {
	    usage();
	    System.exit(1);
	}
	String region = args[2];

	//
	// Actually do the sample
	//
	try {
	    Sample1 sample1 = new Sample1(host,port,region);
	    System.out.println("Date: "+ sample1.getDate());
	    System.out.println("Time: "+ sample1.getTime());
	} catch ( ResourceException e ) {
	    analyseResourceException(e);
	}
    }


//--------------------------------------------------------------------------
// Support methods for the main() function
//--------------------------------------------------------------------------
    /**
     * Read a set of arguments from stdin.
     * <p>This is used if no arguments are supplied on the command line
     * @throws java.io.IOException if the stream closes.
     */
    private static String[] getArgsFromStdIn() throws IOException {
	String[] args = new String[3];
	InputStreamReader isr = new InputStreamReader(System.in);
	BufferedReader br = new BufferedReader(isr);
	System.out.println("Name Service Host (e.g. localhost):");
	args[0] = br.readLine();
	System.out.println("Name Service Port (e.g. 7865):");
	args[1] = br.readLine();
	System.out.println("Region Name (e.g. mtp/mtpsys01):");
	args[2] = br.readLine();
	return args;
    }


    /**
     * Display a usage message.
     */
    private static void usage() {
	System.out.println("Usage: java Sample1 host port region");
    }

    /**
     * Provide a display of a caught resource exception.
     * <p>This demonstrates how to obtain an ABEND code from the exception if
     * the MTP transaction abended.
     * @param e The exception to be analysed.
     */
    private static void analyseResourceException( ResourceException e ) {
	if ( e instanceof ECITransactionAbendException ) {
	    ECITransactionAbendException tae = (ECITransactionAbendException)e;
	    System.err.println(
		    "Transaction abended with abend code "+tae.getAbendCode());
	} else {
	    String errorCode = e.getErrorCode();
	    System.err.println(e.getMessage()+" - Error code "+errorCode);
	}
    }
}

